import logging
from typing import *
from .scpi_resource import ScpiResource
from .scpi_type_base import ScpiTypes, ScpiTypeBase
from .scpi_transport_base import ScpiTransports
from .scpi_types import ScpiEvent
from .scpi_exceptions import ScpiErrorException

logger = logging.getLogger(__name__)

class ScpiDevice(object):
    resource: Optional[ScpiResource] = None
    transport: Optional[ScpiTransports] = None

    def __init__(self, resource: ScpiResource, **kwargs) -> None:
        self.resource = resource
        self.transport = resource.transport(resource.address, **kwargs)
        logger.info(f'open [{resource.transport._transport_class}]: {resource.address}')

    def is_open(self):
        return True if self.transport is not None else False

    def close(self) -> None:
        if self.transport is not None:
            logger.info(f'close [{self.resource.transport._transport_class}]: {self.resource.address}')
            self.transport.close()
            self.transport = None

    def reset(self) -> None:
        self.transport.reset()

    def execute(self, header: str, param: Optional[ScpiTypes] = None,
                result: Optional[Type[ScpiTypes]] = type(None)) -> ScpiTypes:
        header = header.encode(ScpiTypeBase._encoding)
        command = header
        response = None

        if param is not None:
            command = b' '.join([header, param.compose()])

        try:
            self.transport.writeline(command + ScpiTypeBase._newline)

            if result is not type(None):
                response = result.parse(self.transport)

            logger.info(
                f'exec [{type(param).__name__} -> {result.__name__}]: ' +
                f'{str(header)[1:]} ({str(param)}) -> ({str(response)})')
        except Exception as e:
            logger.error(
                f'exec [{type(param).__name__} -> {result.__name__}]: ' +
                f'{str(header)[1:]} ({str(param)}) -> ({str(response)}) [{repr(e)}]')

            # Re-Raise Exception
            raise e

        return response

    def check_error(self) -> Optional[ScpiEvent]:
        """Checks if an error is available in the device's SCPI error queue and returns it. Returns None otherwise"""
        scpi_error = self.execute('SYST:ERR:NEXT?', result=ScpiEvent)
        if scpi_error.code != 0:
            return scpi_error
        return None

    def raise_error(self) -> None:
        """Raise a ScpiErrorException when an scpi error is in the device's SCPI error queue"""
        scpi_error = self.check_error()
        if scpi_error is not None:
            raise ScpiErrorException(scpi_error)

    def __del__(self):
        self.close()

    def __enter__(self) -> 'ScpiDevice':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()


import abc
from typing import *
from typing import BinaryIO


class ScpiTypeBase(object):
    """Base class for SCPI types"""
    __metaclass__ = abc.ABCMeta

    _encoding: str = 'utf-8'
    _newline: bytes = b'\r\n'

    @abc.abstractmethod
    def __str__(self) -> str:
        """A textual representation of the object"""
        raise NotImplementedError("Please Implement this method")

    @abc.abstractmethod
    def compose(self) -> bytes:
        """Converts the given data into SCPI "Device listening" format"""
        raise NotImplementedError("Please Implement this method")

    @classmethod
    @abc.abstractmethod
    def parse(cls, stream: BinaryIO):
        """Parses SCPI "Device talking Elements" into an internal representation"""
        raise NotImplementedError("Please Implement this method")


# TypeVar for type hinting subclasses of ScpiTypeBase
ScpiTypes = TypeVar('ScpiTypes', bound=ScpiTypeBase)



import logging
import re
from typing import *
from typing import BinaryIO
from twopilabs.utils import usbcdc
from .scpi_transport_base import ScpiTransportBase
from .scpi_resource import ScpiResource
from .scpi_exceptions import ScpiTransportException


class ScpiUsbCdcTransport(ScpiTransportBase):
    _transport_class = 'ScpiUsbCdcTransport'
    _transport_info = 'USB CDC SCPI Transport'
    _transport_type = 'USB CDC'

    @classmethod
    def discover(cls, usb_vid: Optional[int] = None, usb_pid: Optional[int] = None) -> List[ScpiResource]:
        return [ScpiResource(transport=ScpiUsbCdcTransport,
                             location=device.location,
                             address=device.address,
                             name=None,
                             manufacturer=device.manufacturer,
                             model=device.product,
                             serialnum=device.serial_number,
                             info=device
                             ) for device in usbcdc.UsbCdcDevice.list_devices(usb_vid=usb_vid, usb_pid=usb_pid)]

    @classmethod
    def from_resource_name(cls, resource_name: str) -> Optional[ScpiResource]:
        m = re.match((
            r'^(?P<prefix>(?P<type>CDC)(?P<board>\d+)?)'
            r'::(?P<vendor_id>((0x[0-9A-Fa-f]+)|([0-9]+)))'
            r'::(?P<product_id>((0x[0-9A-Fa-f]+)|([0-9]+)))'
            r'(::(?P<serial_number>[^\s:]+)?)?'
            r'(::(?P<interface_number>((0x[0-9A-Fa-f]+)|([0-9]+))))?'
            r'(::(?P<suffix>INSTR))?$'
        ), resource_name, re.IGNORECASE)

        if m is None:
            # Does not match the regex
            return None

        groupdict = m.groupdict()

        return ScpiResource(
            transport=ScpiUsbCdcTransport,
            address=usbcdc.UsbCdcDeviceAddress(
                vid=int(groupdict['vendor_id'], 0),
                pid=int(groupdict['product_id'], 0),
                serial_number=groupdict['serial_number'],
                interface_number=int(groupdict['interface_number'], 0) if groupdict['interface_number'] is not None else None
            )
        )

    @classmethod
    def to_resource_name(cls, resource: ScpiResource) -> str:
        address = resource.address
        return f'CDC::0x{address.vid:04x}::0x{address.pid:04x}::' \
               + (f'{address.serial_number}' if address.serial_number is not None else '') \
               + (f'::{address.interface_number}' if address.interface_number is not None else '') \
               + '::INSTR'

    def __init__(self, address: usbcdc.UsbCdcDeviceAddress, timeout: float = 5, **kwargs):
        super().__init__(**kwargs)
        self._logger = logging.getLogger(__name__)
        try:
            device = usbcdc.UsbCdcDevice(address=address, timeout=timeout, **kwargs)
        except usbcdc.UsbCdcException as msg:
            raise ScpiTransportException(msg) from msg

        self._io = cast(BinaryIO, device)



from .scpi_device import ScpiDevice
from .scpi_resource import ScpiResource
from .scpi_types import *

# Import transports, silently fail e.g. if there are missing dependencies
try:
    from .scpi_transport_serial import ScpiSerialTransport
except:
    pass

try:
    from .scpi_transport_tcpip import ScpiTcpIpTransport
except:
    pass

try:
    from .scpi_transport_usbtmc import ScpiUsbTmcTransport
except:
    pass

try:
    from .scpi_transport_usbcdc import ScpiUsbCdcTransport
except:
    pass


from .scpi_types import ScpiEvent


class ScpiException(Exception):
    """This class represents a generic SCPI Error"""
    pass


class ScpiErrorException(ScpiException):
    """This class represents a SCPI Error found in the SYST:ERR queue of the device"""
    def __init__(self, scpi_error: ScpiEvent):
        self.scpi_error = scpi_error
        super().__init__(scpi_error)


class ScpiTransportException(Exception):
    pass



import serial
import logging
import re
from typing import *
from typing import BinaryIO
from .scpi_transport_base import ScpiTransportBase
from .scpi_resource import ScpiResource
from .scpi_exceptions import ScpiTransportException

class ScpiSerialTransport(ScpiTransportBase):
    _transport_class = 'ScpiSerialTransport'
    _transport_info = 'Serial SCPI Transport'
    _transport_type = 'Serial'

    @classmethod
    def discover(cls, usb_vid: Optional[int] = None, usb_pid: Optional[int] = None) -> List[ScpiResource]:
        import serial.tools.list_ports

        return [ScpiResource(transport=ScpiSerialTransport,
                             location=f'{"usb" if port.pid is not None else "serial"}' + f':{port.location}' if port.location is not None else '',
                             address=port.device,
                             name=port.description,
                             manufacturer=port.manufacturer if port.manufacturer is not None else None,
                             model=port.product if port.product is not None else None,
                             serialnum=port.serial_number if port.serial_number is not None else None,
                             info=port
                             ) for port in serial.tools.list_ports.comports()
                if ((usb_vid is None) or (usb_vid == port.vid))
                and ((usb_pid is None) or (usb_pid == port.pid))]

    @classmethod
    def from_resource_name(cls, resource_name: str) -> Optional[ScpiResource]:
        m = re.match((
            r'^(?P<prefix>(?P<type>ASRL)(?P<board>[^\s:]+))'
            r'(::(?P<suffix>INSTR))?$'
        ), resource_name, re.IGNORECASE)

        if m is None:
            # Does not match the regex
            return None

        groupdict = m.groupdict()

        return ScpiResource(
            transport=ScpiSerialTransport,
            address=f'COM{int(groupdict["board"]):d}' if groupdict['board'].isnumeric() else groupdict['board']
        )

    @classmethod
    def to_resource_name(cls, resource: ScpiResource) -> str:
        address = resource.address.partition('COM')[2] if resource.address.startswith('COM') else resource.address
        return f'ASRL{address}::INSTR'

    def __init__(self, port: str, timeout: float = 5, **kwargs):
        super().__init__(**kwargs)
        self._logger = logging.getLogger(__name__)
        try:
            port = serial.Serial(port=port, timeout=timeout, exclusive=True, **kwargs)
        except serial.SerialException as msg:
            raise ScpiTransportException(msg) from msg

        self._io = cast(BinaryIO, port)

        port.reset_input_buffer()
        port.reset_output_buffer()
        port.timeout = 0
        while port.read() != b'':
            pass
        port.timeout = timeout
